using System;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace MyDirect3DProject
{
	public class MainClass : Form
	{
		private Device device;
		private Effect myEffect;
		private VertexDeclaration decl;
		private Mesh myMesh;
		private Texture myTexture;
		private float rotationX;
		private float rotationY;
		private float specularExponent = 30.0f;
		
		public MainClass()
		{
			this.ClientSize = new System.Drawing.Size(640, 480);
			this.Text = "Direct3D Project";
			this.SetStyle(ControlStyles.AllPaintingInWmPaint|ControlStyles.Opaque, true);
		}
		
		public bool InitializeGraphics()
		{
			try {
				PresentParameters presentParams = new PresentParameters();
				presentParams.Windowed = true;
				presentParams.SwapEffect = SwapEffect.Discard;
				presentParams.EnableAutoDepthStencil = true;			
				presentParams.AutoDepthStencilFormat = DepthFormat.D16;
				
				device = new Device(0, DeviceType.Hardware, this,
				                    CreateFlags.HardwareVertexProcessing, presentParams);
				
				device.DeviceLost     += new EventHandler(this.InvalidateDeviceObjects);
				device.DeviceReset    += new EventHandler(this.RestoreDeviceObjects);
				device.Disposing      += new EventHandler(this.DeleteDeviceObjects);
				device.DeviceResizing += new CancelEventHandler(this.EnvironmentResizing);
				
				myEffect = Effect.FromFile(device, @"..\..\shader.fx", null, ShaderFlags.None, null);
				myEffect.Technique = "dx9";
				
				myTexture = TextureLoader.FromFile(device, @"..\..\texture.dds");

				myMesh = Mesh.FromFile(@"..\..\mesh.x", MeshFlags.Managed, device);
				if(myMesh.VertexBuffer.Description.VertexFormat != CustomVertex.PositionNormalTextured.Format)
					throw new Exception("Wrong vertex format used in mesh file.");
				
				VertexElement[] elements = new VertexElement[]
				{
					new VertexElement(0, 0, DeclarationType.Float3, DeclarationMethod.Default, DeclarationUsage.Position, 0),
					new VertexElement(0, 3*4, DeclarationType.Float3, DeclarationMethod.Default, DeclarationUsage.Normal, 0),
					new VertexElement(0, (3+3)*4, DeclarationType.Float2, DeclarationMethod.Default, DeclarationUsage.TextureCoordinate, 0),
					VertexElement.VertexDeclarationEnd
				};
				decl = new VertexDeclaration(device, elements);
			
				return true;
			} catch (DirectXException) {
				return false;
			}			
		}
		
		protected virtual void InvalidateDeviceObjects(object sender, EventArgs e)
		{
		}
		
		protected virtual void RestoreDeviceObjects(object sender, EventArgs e)
		{
		}
		
		protected virtual void DeleteDeviceObjects(object sender, EventArgs e)
		{	
		}
		
		protected virtual void EnvironmentResizing(object sender, CancelEventArgs e)
		{
		}
		
		protected virtual void FrameMove()
		{
			// TODO : Frame movement
		}
		
		protected virtual void Render()
		{
			if (device != null) {
				device.Clear(ClearFlags.Target|ClearFlags.ZBuffer, Color.Blue, 1.0f, 0);
				device.BeginScene();
								
				device.VertexDeclaration = decl;				
				device.SetTexture(0, myTexture);
								
				Matrix proj = Matrix.PerspectiveFovLH(0.8f, Width/(float)Height, 0.1f, 50.0f);
				Matrix view = Matrix.LookAtLH(new Vector3(0.0f, 0.0f, -2.0f), new Vector3(), new Vector3(0.0f, 1.0f, 0.0f));
				Matrix world = Matrix.RotationX(rotationX)*Matrix.RotationY(rotationY);
				Matrix worldT = world;
				worldT.Transpose(worldT);
				Matrix worldIT = worldT;
				worldIT.Invert();
				Matrix viewI = view;
				viewI.Invert();

				myEffect.SetValue("W", world);
				myEffect.SetValue("WVP", world*view*proj);
				myEffect.SetValue("WIT", worldIT);
				myEffect.SetValue("VI", viewI);
				myEffect.SetValue("LightPos", new Vector4(-5f, 7f, -4f, 1.0f));				
				myEffect.SetValue("SpecularColor", new Vector4(1.0f, 1.0f, 0.0f, 1.0f));
				myEffect.SetValue("SpecularExponent", specularExponent);

				int numPasses = myEffect.Begin(0);
				for(int pass = 0; pass < numPasses; pass++)
				{
					myEffect.Pass(pass);
					myMesh.DrawSubset(0);
				}
				myEffect.End();
				
				device.EndScene();
				device.Present();
			}
		}
		
		public void Run()
		{
			while (Created) {
				FrameMove();	
				Render();
				Application.DoEvents();
			}
		}
		
		protected override void OnPaint(PaintEventArgs e)
		{
			this.Render();
		}

		protected override void OnKeyDown(KeyEventArgs e)
		{
			base.OnKeyDown(e);
			switch(e.KeyCode)
			{
				case Keys.Up:
					rotationX += 0.1f;
					break;
				case Keys.Down:
	            	rotationX -= 0.1f;
					break;
				case Keys.Left:
	            	rotationY += 0.1f;
					break;
				case Keys.Right:
					rotationY -= 0.1f;
					break;
				case Keys.PageUp:
	            	specularExponent *= 1.1f;
					break;
				case Keys.PageDown:
					specularExponent /= 1.1f;
					break;
			}
		}
		
		protected override void OnKeyPress(KeyPressEventArgs e)
		{
			base.OnKeyPress(e);
			switch((int)e.KeyChar)
			{
				case (int)Keys.Escape:
					this.Close();
					break;
			}
		}
		
		static void Main()
		{
			using (MainClass mainClass = new MainClass()) {
				if (!mainClass.InitializeGraphics()) {
					MessageBox.Show("Error while initializing Direct3D");
					return;
				}
				mainClass.Show();
				mainClass.Run();
			}
		}
	}
}
